; ----------------------------------------------------------------------------
; render_teletext_row(a0:page row 40 bytes, a1:bitplane0, a2:bitplane1, d5:MPAG) -- *a3 *a4
; ----------------------------------------------------------------------------
WORD_OFF=(128*4)

render_teletext_row:

	; byte	|0	|1	|2	|3	|4	|5	|6	|7
	;	|clock	|clock	|frame	|mpag	|mpag	|data	|data	|...

	; word	|0	|2	|4	|6	|8	|10	|12	|14
	;	[even----------][even----------][even----------][even----------]
	;		[odd-----------][odd-----------][odd-----------][odd-----------]

	; [0,1,2] even, odd, even: 2 bytes clock run-in, 1 byte framing code
	move.w	#%1100110011001100,(a2)+	; clock run-in 0 1
	move.w	#%1100110001100110,(a2)+	; clock run-in 2 3
	move.w	#%0111111000011000,(a2)+	; framing code 4 5

	moveq	#0,d0
	move.w	d0,(a2)+			; even at 8
	move.l	d0,(a1)+
	move.w	d0,(a1)+			; odd  at 6

	lea	pixeldata,a3
	add.w	#WORD_OFF*3,a3		; pixels for column 3

	; [3,4] odd, even: MPAG
	lea	nibble_to_ham_table,a4
	move.w	d5,d0
	and.w	#$0f,d0
	move.b	(a4,d0.w),d0		; ham-encoded
	add.w	d0,d0
	add.w	d0,d0
	move.l	(a3,d0.w),(a1)+		; odd byte longword
	add.w	#WORD_OFF,a3

	move.w	d5,d0
	lsr.w	#4,d0
	move.b	(a4,d0.w),d0		; ham-encoded
	add.w	d0,d0
	add.w	d0,d0
	move.l	(a3,d0.w),(a2)+		; even byte longword
	add.w	#WORD_OFF,a3


	; [5,6, ..., 43,44]: character data
	move.w	#20-1,d1
.cols	moveq	#0,d0
	move.b	(a0)+,d0
	add.w	d0,d0
	add.w	d0,d0
	move.l	(a3,d0.w),(a1)+		; odd
	add.w	#WORD_OFF,a3

	moveq	#0,d0
	move.b	(a0)+,d0
	add.w	d0,d0
	add.w	d0,d0
	move.l	(a3,d0.w),(a2)+		; even
	add.w	#WORD_OFF,a3

	dbf	d1,.cols

	move.w	#0,(a1)+

	rts

nibble_to_ham_table:
	dc.b	%00010101	; ham(%0000) = %00010101 | parity(%00010101) = %00010101 | matches original hammed? true
	dc.b	%00000010	; ham(%0001) = %00000010 | parity(%00000010) = %00000010 | matches original hammed? true
	dc.b	%01001001	; ham(%0010) = %01001001 | parity(%01001001) = %01001001 | matches original hammed? true
	dc.b	%01011110	; ham(%0011) = %01011110 | parity(%01011110) = %01011110 | matches original hammed? true
	dc.b	%01100100	; ham(%0100) = %01100100 | parity(%01100100) = %01100100 | matches original hammed? true
	dc.b	%01110011	; ham(%0101) = %01110011 | parity(%01110011) = %01110011 | matches original hammed? true
	dc.b	%00111000	; ham(%0110) = %00111000 | parity(%00111000) = %00111000 | matches original hammed? true
	dc.b	%00101111	; ham(%0111) = %00101111 | parity(%00101111) = %00101111 | matches original hammed? true
	dc.b	%01010000	; ham(%1000) = %11010000 | parity(%01010000) = %11010000 | matches original hammed? true
	dc.b	%01000111	; ham(%1001) = %11000111 | parity(%01000111) = %11000111 | matches original hammed? true
	dc.b	%00001100	; ham(%1010) = %10001100 | parity(%00001100) = %10001100 | matches original hammed? true
	dc.b	%00011011	; ham(%1011) = %10011011 | parity(%00011011) = %10011011 | matches original hammed? true
	dc.b	%00100001	; ham(%1100) = %10100001 | parity(%00100001) = %10100001 | matches original hammed? true
	dc.b	%00110110	; ham(%1101) = %10110110 | parity(%00110110) = %10110110 | matches original hammed? true
	dc.b	%01111101	; ham(%1110) = %11111101 | parity(%01111101) = %11111101 | matches original hammed? true
	dc.b	%01101010	; ham(%1111) = %11101010 | parity(%01101010) = %11101010 | matches original hammed? true

; ----------------------------------------------------------------------------
; set_page_header(a0:header, a1:page)
; ----------------------------------------------------------------------------
; Idea was to take page number and flags as parameters -- later...
;
; For now, removes highest bit -- will be added again by the pre-parity-encoded
; byte bitmaps: parity(hamming(x)) == hamming(x)
;
set_page_header:
	move.l	(a0)+,d0
	and.l	#$7f7f7f7f,d0
	move.l	d0,(a1)+
	move.l	(a0)+,d0
	and.l	#$7f7f7f7f,d0
	move.l	d0,(a1)+
	rts
