; ---------------------------------------------------------------------------
;	Teletest
; ---------------------------------------------------------------------------
;	Losso/AttentionWhore <losso@heckmeck.de>
; ---------------------------------------------------------------------------
PAGE_ROWS	= 24
DISPLAY_ROWS	= 14		; ~ half page shown every tick
HIRES_W		= 768		;  96 b
HIRES_WB	= (HIRES_W/8)
TEXT_BPL_ON	= $0901
TEXT_BPL_OFF	= (TEXT_BPL_ON+DISPLAY_ROWS*$0100)
PREVIEW_BG_ON	= $2207
PREVIEW_BG_OFF	= $3507
REPEAT_BPL_ON	= $4001
REPEAT_BPL_OFF	= (REPEAT_BPL_ON+DISPLAY_ROWS*$0100)
DEBUG_DUMP_ON	= $f307
DEBUG_DUMP_OFF	= (DEBUG_DUMP_ON+DISPLAY_ROWS*4*$0100)&$ffff
MOUSE_MAXX	= 40*2-1
MOUSE_MAXY	= 24*3-1
HELLO_TICKS	= (25*10)
ENABLE_LOG	= 1
; ---------------------------------------------------------------------------

	include	macros.asm

	section code,code

	include	startup.asm	; Takeover/restore

; ---------------------------------------------------------------------------
; Before system takeover; return non-zero on error
; ---------------------------------------------------------------------------

Precalc:
	lea	vars(pc),a5
	lea	$dff000,a6
	bsr	teletext_init
	bsr	testpage_init
	bsr	init_random
	moveq	#0,d0		; Precalc OK
	rts

; ---------------------------------------------------------------------------
; Main thread after system takeover (a5: vars, a6: custom)
; ---------------------------------------------------------------------------

Main:
	lea	$dff000,a6
	lea	vars,a5
	move.l	#copperlist,$80(a6)
	bsr	readkey
	clr.w	_key(a5)
	clr.w	_lastkey(a5)
	st	_running(a5)
.loop	;
	; Main thread
	;
	move.l	_txt_page(a5),a0
	lea	preview_area,a1
	bsr	render_preview
	tst.w	_running(a5)
	bne.b	.loop
	;
	; Restore system
	;
	rts

; ---------------------------------------------------------------------------
; Exit code (system is restored)
; ---------------------------------------------------------------------------

Exit:
	moveq	#0,d0
	rts

; ---------------------------------------------------------------------------
; Interrupt
; ---------------------------------------------------------------------------

IRQ:
	movem.l	d0-a6,-(a7)
	lea	vars(pc),a5
	lea	$dff000,a6

	btst	#5,$1f(a6)		; VERTB in INTREQR
	bne	VBI_IRQ

	addq.w	#1,_ticks(a5)
	btst	#6,$bfe001
	bne.b	.nolmb
	btst	#2,$dff016
	bne.b	.nolmb
	clr.w	_running(a5)
.nolmb	bsr	render_step
	bsr	update_mouse

	lea	$dff09c,a6		; INTREQ
	move.w	#%0000000000010000,d0	; clear COPER
;		  ab-------cdefg--
;	a: SET/CLR Bit
;	b: Master Bit
;	c: Blitter Int
;	d: Vert Blank Int
;	e: Copper Int
;	f: IO Ports/Timers
;	g: Software Int
	move.w	d0,(a6)
	move.w	d0,(a6)			; twice to avoid a4k hw bug
	movem.l	(a7)+,d0-a6
	rte

VBI_IRQ:
	tst.w	_log_vbi(a5)
	beq	.no_vbi_log
	move.w	$4(a6),d0		; VPOSR
	DEBUGLOG 'VBI called!         (VPOSR=$%04x)',d0
.no_vbi_log

	bsr	update_vbstop

	lea	$dff09c,a6		; INTREQ
	move.w	#%0000000000100000,d0	; clear VERTB
;		  ab-------cdefg--
;	a: SET/CLR Bit
;	b: Master Bit
;	c: Blitter Int
;	d: Vert Blank Int
;	e: Copper Int
;	f: IO Ports/Timers
;	g: Software Int
	move.w	d0,(a6)
	move.w	d0,(a6)			; twice to avoid a4k hw bug
	movem.l	(a7)+,d0-a6
	rte

; As suggested by ross:
;
; > ... you should remove the writing of VBSTRT from the copper list and do it during the IRQ3 of the vertical blank.
; > The value to write is $137+LOF bit (which you read from VPOSR).
;
update_vbstop:
	tst.w	_dyn_vbi(a5)
	beq.b	.vbstop_in_copperlist
	move.w	#$01fe,txtvbstrt	; disable VBSTRT in copperlist
	move.w	txtvbstrt+2,d0		; but use the value from there
	btst	#7,$4(a6)		; bit 15 of VPOSR = LOF
	beq.b	.no_add_1
	addq.w	#1,d0
.no_add_1
	move.w	d0,$1cc(a6)		; set VBSTRT directly
	tst.w	_log_vbi(a5)
	bne.b	.do_log
	rts
.do_log
	DEBUGLOG 'VBSTRT set to $%04x (VPOSR=$%04x)',d0,$4(a6)
	rts

.vbstop_in_copperlist
	move.w	#$01cc,txtvbstrt
	rts

; ---------------------------------------------------------------------------
; Code parts and data
; ---------------------------------------------------------------------------

	include	vars.asm
	include	input.asm
	include	teletext.asm
	include	testpages.asm
	include	debug-log.asm
	include	encoding.asm
	include	util.asm
	include	data.asm

; ---------------------------------------------------------------------------
; Chip
; ---------------------------------------------------------------------------

	section	data_c,data_c

	include	chipram.asm

; ---------------------------------------------------------------------------
