; ----------------------------------------------------------------------------
; System takeover/restore
; ----------------------------------------------------------------------------

INTENASET	= %1100000000110000
;		   ab-------cdefg--
;	a: SET/CLR Bit
;	b: Master Bit
;	c: Blitter Int
;	d: Vert Blank Int
;	e: Copper Int
;	f: IO Ports/Timers
;	g: Software Int

DMASET		= %1000001111100000
;		   a----bcdefghi--j
;	a: SET/CLR Bit
;	b: Blitter Priority
;	c: Enable DMA
;	d: Bit Plane DMA
;	e: Copper DMA
;	f: Blitter DMA
;	g: Sprite DMA
;	h: Disk DMA
;	i..j: Audio Channel 0-3		; abcdefghijklmnop

stingray_ministartup:
	;
	; Startup/restore code from StingRay's mini startup, modified
	; - Precalc, Exit hooks
	; - Use "vars" in a5
	;
	; Any bugs introduced by me, probably. Original version:
	; https://github.com/MK1Roxxor/MiniStartup/blob/master/startup.i
	;
	;
	movem.l	d0-a6,-(a7)
	bsr	Precalc
	tst.w	d0
	bne	.restore_regs_and_exit

	move.l	4.w,a6
	lea	vars(pc),a5
	lea	gfxname(pc),a1
	moveq	#0,d0
	jsr	-552(a6)		; OpenLibrary()
	move.l	d0,_gfxbase(a5)
	beq.b	.restore
	move.l	d0,a6
	move.l	34(a6),_oldview(a5)
	sub.l	a1,a1
	bsr.w	.doview
	move.l	$26(a6),_oldcop1(a5)	; Store old CL 1
	move.l	$32(a6),_oldcop2(a5)	; Store old CL 2
	bsr	.getvbr
	move.l	d0,_vbrptr(a5)
	move.l	d0,a0
	lea	$dff000,a6
	move.w	$10(a6),_adk(a5)	; Store old ADKCON
	move.w	$1c(a6),_intenaval(a5)	; Store old INTENA
	move.w	$02(a6),_dma(a5)	; Store old DMA
	move.w	#$7fff,d0
	bsr	.waitraster
	move.w	d0,$9a(a6)		; Disable Interrupts
	move.w	d0,$96(a6)		; Clear all DMA channels
	move.w	d0,$9c(a6)		; Clear all INT requests
	move.l	$6c(a0),_oldvbi(a5)
	lea	IRQ(pc),a1
	move.l	a1,$6c(a0)
	move.w	#INTENASET,$9a(a6)	; INTENA
	move.w	#DMASET,$96(a6)		; DMACON
	bsr	Main
.restore
	lea	vars(pc),a5
	lea	$dff000,a6
	move.w	#$8000,d0
	or.w	d0,_intenaval(a5)	; SET/CLR-Bit to 1
	or.w	d0,_dma(a5)		; SET/CLR-Bit to 1
	or.w	d0,_adk(a5)		; SET/CLR-Bit to 1
	subq.w	#1,d0
	bsr	.waitraster
	move.w	d0,$9a(a6)		; Clear all INT bits
	move.w	d0,$96(a6)		; Clear all DMA channels
	move.w	d0,$9c(a6)		; Clear all INT requests
	move.l	_vbrptr(a5),a0
	move.l	_oldvbi(a5),$6c(a0)
	move.l	_oldcop1(a5),$80(a6)	; Restore old CL 1
	move.l	_oldcop2(a5),$84(a6)	; Restore old CL 2
	move.w	d0,$88(a6)		; start copper1
	move.w	_intenaval(a5),$9a(a6)	; Restore INTENA
	move.w	_dma(a5),$96(a6)	; Restore DMAcon
	move.w	_adk(a5),$9e(a6)	; Restore ADKcon
	move.l	_gfxbase(a5),a6
	move.l	_oldview(a5),a1		; restore old viewport
	bsr.b	.doview
	move.l	a6,a1
	move.l	4.w,a6
	jsr	-414(a6)		; Closelibrary()
.restore_regs_and_exit
	movem.l	(a7)+,d0-a6
	bra	Exit

.doview	jsr	-222(a6)		; LoadView()
	jsr	-270(a6)		; WaitTOF()
	jmp	-270(a6)

.getvbr	move.l	a5,-(a7)
	moveq	#0,d0			; default at $0
	move.l	4.w,a6
	btst	#0,296+1(a6)		; 68010+?
	beq.b	.is68k			; nope.
	lea	.getit(pc),a5
	jsr	-30(a6)			; SuperVisor()
.is68k	move.l	(a7)+,a5
	rts
.getit	dc.w	$4e7a,$0801		; = movec vbr,d0
	rte				; back to user state code

.waitraster
	move.l	d0,-(a7)
.loop	move.l	$dff004,d0
	and.l	#$1ff00,d0
	cmp.l	#303<<8,d0
	bne.b	.loop
	move.l	(a7)+,d0
	rts

gfxname:
	dc.b	'graphics.library',0,0
