; ---------------------------------------------------------------------------
;
;	Teletext encoding stuff
;
; ---------------------------------------------------------------------------
;
; - Page size used: 40 * 24 (Level 1 Teletext as per spec [1])
; - 25 fps is assumed -- we cannot transmit a whole page in 1 frame anyway [2]
; - We transmit 14 lines per frame
; - TV lingo vs Amiga
;   - TV:    1 frame = 2 interlaced fields (lines 0 to 625)
;   - Amiga: 1 frame = what is displayed in 1/50 sec (VPOS 0 to 285-something)
; - etc. yadda yadda
;
; [1] https://www.etsi.org/deliver/etsi_i_ets/300700_300799/300706/01_60/ets_300706e01p.pdf
; [2] Teletext spec 4.1
;     > Lines 6 to 22 and 318 to 335 may be used to carry Teletext data packets
;     6..22 = 17 lines, 318..335 = 18 lines
;
; ---------------------------------------------------------------------------

COL0_BLACK	= 0
COL1_RED	= 1
COL2_GREEN	= 2
COL3_YELLOW	= 3
COL4_BLUE	= 4
COL5_MAGENTA	= 5
COL6_CYAN	= 6
COL7_WHITE	= 7

SET_TXT_COL0_BLACK	= COL0_BLACK
SET_TXT_COL1_RED	= COL1_RED
SET_TXT_COL2_GREEN	= COL2_GREEN
SET_TXT_COL3_YELLOW	= COL3_YELLOW
SET_TXT_COL4_BLUE	= COL4_BLUE
SET_TXT_COL5_MAGENTA	= COL5_MAGENTA
SET_TXT_COL6_CYAN	= COL6_CYAN
SET_TXT_COL7_WHITE	= COL7_WHITE
SET_GFX_COL0_BLACK	= $10
SET_GFX_COL1_RED	= $11
SET_GFX_COL2_GREEN	= $12
SET_GFX_COL3_YELLOW	= $13
SET_GFX_COL4_BLUE	= $14
SET_GFX_COL5_MAGENTA	= $15
SET_GFX_COL6_CYAN	= $16
SET_GFX_COL7_WHITE	= $17

SET_NEW_BG	= $1d

teletext_init:
	addq.w	#1,_ticks(a5)

	move.l	#preview_area,d0
	move.w	d0,previewbpl+6
	move.w	d0,previewbpl+6+8
	swap	d0
	move.w	d0,previewbpl+2
	move.w	d0,previewbpl+2+8

	move.l	#data_bitmap_a,d0
	move.w	d0,debugdumppls+6
	swap	d0
	move.w	d0,debugdumppls+2
	swap	d0
	add.l	#HIRES_WB,d0
	move.w	d0,debugdumppls+6+8
	swap	d0
	move.w	d0,debugdumppls+2+8

	move.l	#preview_start,d0
	move.w	d0,repeat_skip+6
	swap	d0
	move.w	d0,repeat_skip+2

	move.l	#back2black,d0
	move.w	d0,dump_skip+6
	swap	d0
	move.w	d0,dump_skip+2

	move.l	#coppersprites,d0
	lea	coppersprites+6,a1
	move.w	d0,(a1)
	move.w	d0,8*1(a1)
	move.w	d0,8*2(a1)
	move.w	d0,8*3(a1)
	move.w	d0,8*4(a1)
	move.w	d0,8*5(a1)
	move.w	d0,8*6(a1)
	move.w	d0,8*7(a1)
	swap	d0
	subq	#4,a1
	move.w	d0,(a1)
	move.w	d0,8*1(a1)
	move.w	d0,8*2(a1)
	move.w	d0,8*3(a1)
	move.w	d0,8*4(a1)
	move.w	d0,8*5(a1)
	move.w	d0,8*6(a1)
	move.w	d0,8*7(a1)

	move.l	#font_norm,_font(a5)
	move.l	#font_norm+128*8,_fontgfx(a5)

	clr.w	_current_testpage(a5)
	move.w	#1,_dyn_vbi(a5)
	clr.w	_log_vbi(a5)
	clr.w	_debug_log_pause(a5)
	clr.w	_txt_slowdown(a5)
	clr.w	_txt_slowdown_to_go(a5)

	move.l	#startscreen,_txt_page(a5)
	move.l	#testpage_renderer,_txt_renderer(a5)

		; YYYYYMMM
	move.w	#%00000001,_txt_page_mag(a5)	; initial MAG with Y=0
	move.w	#%00000001,_txt_filler_mag(a5)	; use MPAG 1xx, keep Y=0

	rts

; ---------------------------------------------------------------------------
;
; Render mode: Display, rendering and encoding interleaved
;
; Data bitmaps A, B: encoded Teletext packets to be displayed during VBI
; Page buffer: 40*24 bytes for display
;
; Let's hope neither page rendering nor encoding need more than 1 frame...
;
; Frame | Display             | Render             | Encode packets
;     0 | Upper data bitmap A | Update page buffer |
;     1 | Lower data bitmap A |                    | Page buffer -> data bitmap B
;     2 | Upper data bitmap B | Update page buffer |
;     3 | Lower data bitmap B |                    | Page buffer -> data bitmap A
;
; ---------------------------------------------------------------------------
; render_step
; ---------------------------------------------------------------------------
render_table:
	dc.l	data_bitmap_a+0,			data_bitmap_b
	dc.l	data_bitmap_a+DISPLAY_ROWS*HIRES_WB*2,	data_bitmap_b
	dc.l	data_bitmap_b+0,			data_bitmap_a
	dc.l	data_bitmap_b+DISPLAY_ROWS*HIRES_WB*2,	data_bitmap_a
render_step:
	;
	; update pointers according to table
	;
	move.w	_ticks(a5),d0
	addq	#1,d0			; compensate for long/short field or some shit
	and.w	#3,d0
	move.w	d0,d2
	lsl.w	#3,d0
	lea	render_table(pc),a0
	add.w	d0,a0
	movem.l	(a0)+,d0/d1
	move.l	d0,_txt_display_bpl(a5)
	move.l	d1,_txt_encode_bpl(a5)
	;
	; set display bpls
	;
	lea	txtbpls+2,a1
	move.w	d0,4(a1)
	swap	d0
	move.w	d0,(a1)
	swap	d0
	add.l	#HIRES_WB,d0		; second bpl
	move.w	d0,8+4(a1)
	swap	d0
	move.w	d0,8(a1)
	;
	; also set preview bpls
	;
	lea	txtbpls+2,a0
	lea	txtbpls2+2,a1
	move.w	(a0),(a1)		; BPL1H
	move.w	4(a0),4(a1)		; BPL1L
	move.w	8(a0),8(a1)		; BPL2H
	move.w	12(a0),12(a1)		; BPL2L
	;
	; finally, either render page or encode
	;
	btst	#0,d2			; d2 was ticks % 3
	bne.b	.render
	bsr	readkey			; readkey only every other frame so we don't lose keys
	move.l	_txt_renderer(a5),a0
	cmp.l	#0,a0			; no renderer set?
	bne.b	.has_renderer
	rts
.has_renderer
	tst.w	_debug_time(a5)
	bne.b	.render_with_debug
	jsr	(a0)
	clr.w	_mouse_dx(a5)
	clr.w	_mouse_dy(a5)
	rts
.render_with_debug
	move.w	#$0ff,$180(a6)
	jsr	(a0)
	rept 4
	move.w	#$0ff,$180(a6)
	move.w	#$088,$180(a6)
	endr
	move.w	#$666,$180(a6)
	clr.w	_mouse_dx(a5)
	clr.w	_mouse_dy(a5)
	rts
	;
	; encode
	;
.render
	;
	; fps slowdown?
	;
	move.w	_txt_slowdown_to_go(a5),d0
	beq.b	.no_slowdown
	subq.w	#1,d0
	move.w	d0,_txt_slowdown_to_go(a5)
	bra	send_filler
.no_slowdown
	;
	; set page header for lower two page digits 00
	;
	lea	page_00_header,a0
	move.l	_txt_page(a5),a1
	bsr	set_page_header

	move.l	_txt_page(a5),a0
	move.l	_txt_encode_bpl(a5),a1
	lea	HIRES_WB(a1),a2		; second bpl
	move.w	_txt_page_mag(a5),d5	; initial MPAG with Y=0
	move.w	#PAGE_ROWS-1,d7		; rows
	tst.w	_debug_time(a5)
	bne.b	.debug
.rows	bsr	render_teletext_row
	add.w	#%00001000,d5		; inc Y
	add.w	#HIRES_WB*2-92,a1
	add.w	#HIRES_WB*2-92,a2
	dbf	d7,.rows
	;
	; extra rows: re-use first row as time filling header (xFF)
	;
	move.l	a1,a3
	lea	page_ff_header,a0
	move.l	_txt_page(a5),a1
	bsr	set_page_header
	move.l	a3,a1

	move.w	_txt_filler_mag(a5),d5
	move.w	#(DISPLAY_ROWS*2-PAGE_ROWS)-1,d7
.xtra	move.l	_txt_page(a5),a0	; always re-use first row with page header
	bsr	render_teletext_row
	add.w	#HIRES_WB*2-92,a1
	add.w	#HIRES_WB*2-92,a2
	dbf	d7,.xtra
	;
	; fps slowdown?
	;
	move.w	_txt_slowdown(a5),_txt_slowdown_to_go(a5)

	rts
; ---------------------------------------------------------------------------
.debug
.rows_with_debug
	bsr	render_teletext_row
		; YYYYYMMM
	add.w	#%00001000,d5		; inc Y
	add.w	#HIRES_WB*2-92,a1
	add.w	#HIRES_WB*2-92,a2
	move.w	#$800,$dff180
	dbf	d7,.rows_with_debug
	move.w	#$080,$dff180
	move.w	#$0f0,$dff180
	move.w	#$080,$dff180
	move.w	#$0f0,$dff180
	move.w	#$080,$dff180
	move.w	#$0f0,$dff180
	move.w	#$666,$dff180
	rts

; ---------------------------------------------------------------------------
send_filler:
	move.l	_txt_page(a5),a0
	move.l	_txt_encode_bpl(a5),a1
	lea	HIRES_WB(a1),a2		; second bpl
	move.w	_txt_page_mag(a5),d5	; initial MPAG with Y=0

	move.l	a1,a3
	lea	page_ff_header,a0
	move.l	_txt_page(a5),a1
	bsr	set_page_header
	move.l	a3,a1

	move.w	_txt_filler_mag(a5),d5
	move.w	#(DISPLAY_ROWS*2)-1,d7
.xtra	move.l	_txt_page(a5),a0	; always re-use first row with page header
	bsr	render_teletext_row
	add.w	#HIRES_WB*2-92,a1
	add.w	#HIRES_WB*2-92,a2
	dbf	d7,.xtra
	rts

; ---------------------------------------------------------------------------
; un_parity(a1: buffer)
; ---------------------------------------------------------------------------
; Remove parity bits for bin dumps exported from a Teletext editor - oops, not needed
un_parity:
	move.l	#40*24/4-1,d0
	move.l	#$7f7f7f7f,d1
.loop	and.l	d1,(a1)+
	dbf	d0,.loop
	rts

; ---------------------------------------------------------------------------
; render_preview(a0: teletext page, a1: bitmap)
; ---------------------------------------------------------------------------
;
; Slow as molasses on 68000. This thread has many optimizations ideas:
; https://eab.abime.net/showthread.php?t=60985
;
; Simple preview, not a Teletext emulator. No colors, no blinking :)
;
render_preview:
	moveq	#8,d5		; skip bytes (8 bytes of page header)
	move.w	#24-1,d7
.rows	move.l	_font(a5),a2	; start each row with alphanum font
	move.w	#40-1,d6
.cols	moveq	#0,d0
	move.b	(a0)+,d0	; code
	subq	#1,d5		; skip?
	blt.b	.noskip
	move.b	#' ',d0
.noskip	;
	; $00..$1f = control char
	;
	cmp.w	#$20,d0
	bge.b	.control_check_done
	;
	; $00..$07 = set color + alphanum
	;
	cmp.w	#$07,d0
	bgt.b	.no_set_alpha
	move.l	_font(a5),a2	; set alphanum font
	bra.b	.control_check_done
.no_set_alpha
	;
	; $10..$16 = set color + graphics
	;
	cmp.w	#$17,d0
	bgt.b	.control_check_done
	cmp.w	#$10,d0
	blt.b	.control_check_done
	move.l	_fontgfx(a5),a2	; set graphics font
	;
	; plot
	;
.control_check_done
	lsl.w	#3,d0		; one glyph every 8 bytes
	lea	(a2,d0.w),a3
	move.b	(a3)+,(a1)
	move.b	(a3)+,40*1(a1)
	move.b	(a3)+,40*2(a1)
	move.b	(a3)+,40*3(a1)
	move.b	(a3)+,40*4(a1)
	move.b	(a3),40*5(a1)

	addq	#1,a1
	dbf	d6,.cols
	add.w	#40*5,a1
	dbf	d7,.rows
	rts

; ---------------------------------------------------------------------------
; clear_buffer(a1: buffer)
; ---------------------------------------------------------------------------
clear_buffer:
	move.l	#$20202020,d0
	move.w	#(40*24)/4-1,d1
.clr	move.l	d0,(a1)+
	dbf	d1,.clr
	rts
; ---------------------------------------------------------------------------
copy_buffer:
	move.w	#(40*24)/4-1,d1
.clr	move.l	(a0)+,(a1)+
	dbf	d1,.clr
	rts
; ---------------------------------------------------------------------------
copy_buffer_masked:
	move.w	#(40*24)/4-1,d1
.clr	move.l	(a0)+,d0
	and.l	#$7f7f7f7f,d0
	move.l	d0,(a1)+
	dbf	d1,.clr
	rts
; ---------------------------------------------------------------------------
copy_buffer_rightshift:
	move.w	#24-1,d1
	move.l	a0,a2
.rows	move.l	a2,a0
	move.w	#40-1,d2
.cols	cmp.w	d2,d0
	bge.b	.cp
	move.b	#$20,(a1)+
	bra.b	.next
.cp	move.b	(a0)+,(a1)+
.next	dbf	d2,.cols
	add.w	#40,a2
	dbf	d1,.rows
	rts
; ---------------------------------------------------------------------------
copy_only_colorsetters:
	move.w	#(40*24)-1,d1
.clr	move.b	(a0)+,d0
	cmp.b	#$18,d0
	blt.b	.docp
	move.w	#' ',d0
.docp	move.b	d0,(a1)+
	dbf	d1,.clr
	rts
; ---------------------------------------------------------------------------
copy_faded_with_mapping_in_a2:		; a0 -> a1
	move.w	#24-1,d6
.rows	move.w	#40-1,d5
.cols	moveq	#0,d0
	move.b	(a0)+,d0		; src char
	cmp.b	#$08,d0			; $00..$07 = alpha + col
	bge.b	.notalphacol
	move.b	(a2,d0.w),d0		; lookup fade
	bra.b	.setchar
.notalphacol
	cmp.w	#$18,d0			; $10..$17 = gfx + col
	bge.b	.setchar
	cmp.w	#$10,d0
	blt.b	.setchar
	and.b	#$07,d0
	move.b	(a2,d0.w),d0		; lookup fade
	or.b	#$10,d0
.setchar
	move.b	d0,(a1)+
	dbf	d5,.cols
	dbf	d6,.rows
	rts
; ---------------------------------------------------------------------------
; collect_color_info(a0: page buffer, a1: color buffer) *a0 *a1 *d0 *d1 *d2 *d3
; ---------------------------------------------------------------------------
; Remove all non-color-setter bytes and replace everything else with the
; last color setter character.
;
; Color setter codes:
;
; Bits    Hex Effect
; ------- --- ------
; --T-CCC
; ..0.000 $00 Text mode, foreground black
; ..0.001 $01 Text mode, foreground green
; ..0.010 $02 Text mode, foreground yellow
;             ...
; ..0.111 $07 Text mode, foreground white
; ..1.000 $10 Gfx mode, foreground black
; ..1.001 $11 Gfx mode, foreground green
;             ...
; ..1.111 $17 Gfx mode, foreground white
;
collect_color_info:
	move.w	#PAGE_ROWS-1,d1
.rows	move.w	#$07,d2			; each line starts with alpha white
	move.w	#40-1,d3
.cols	move.b	(a0)+,d0
	and.b	#%11101000,d0		; mask out all bits not used by color setter commands
	bne.b	.no_setter
	move.b	-1(a0),d2		; non-colorsetter bits are zero: update color setter
.no_setter
	move.b	d2,(a1)+
	dbf	d3,.cols
	dbf	d1,.rows
	rts

update_renderbuffer_color_info:
	lea	renderbuffer,a0
	lea	40*24(a0),a1		; assume room for color buffer right after page data
	bra	collect_color_info	; collect_color_info(a0: page buffer, a1: color buffer) *a0 *a1 *d0 *d1 *d2 *d3

; ---------------------------------------------------------------------------
; draw_sixel_sprite(a0: sprite, a1: buf, d0: x, d1: y, d2: color setter command)
; ---------------------------------------------------------------------------
draw_sixel_sprite:
	;
	; record current color setters
	;
	movem.l	d0-d4/a0-a1,-(a7)
	move.l	a1,a0
	lea	40*24(a0),a1		; assume room for color buffer right after page data
	bsr	collect_color_info	; collect_color_info(a0: page buffer, a1: color buffer) *a0 *a1 *d0 *d1 *d2 *d3
	movem.l	(a7)+,d0-d4/a0-a1

	;
	; select source bitmap (out of the 6 shifted ones)
	;
	lea	5*2(a0),a2	; a2 = source bitmap (starts after 5 words for hotx, hoty, bplsize, w-1, h-1)

	sub.w	(a0)+,d0	; x -= hotspot x
	sub.w	(a0)+,d1	; y -= hotspot y
	move.w	(a0)+,d3	; d3 = bitmap size
	;
	; select sprite bitmap from bitmaps
	;
	; by sixel offset from top-left character corner:
	;
	; bitmap 0 = +0 +0  <-- x if even  y % 3 == 0
	; bitmap 1 = +1 +0  <-- x is odd   y % 3 == 0
	; bitmap 2 = +0 +1  <-- x if even  y % 3 == 1
	; bitmap 3 = +1 +1  <-- x is odd   y % 3 == 1
	; bitmap 4 = +0 +2  <-- x if even  y % 3 == 2
	; bitmap 5 = +1 +2  <-- x is odd   y % 3 == 2
	;
	; x odd    --> bitmap start += 1 sprite
	;
	lsr.w	#1,d0			; set C flag if lowest bit was set
	bcc.b	.x0
	add.w	d3,a2			; x odd
.x0	;
	; y % 3 == 1 --> bitmap start += 2 sprites
	; y % 3 == 2 --> bitmap start += 4 sprites
	;
	swap	d1			; clear upper half of d1 for division
	clr.w	d1
	swap	d1

	divu	#3,d1
	swap	d1			; y % 3
	tst.w	d1			; y % 3 == 0
	beq.b	.startfound
	add.w	d3,a2			; y % 3 == 1 or 2
	add.w	d3,a2
	btst	#0,d1
	bne.b	.startfound
	add.w	d3,a2			; y % 3 == 2
	add.w	d3,a2
.startfound
	;
	; set buffer target
	;
	swap	d1			; d1 = y/3
	mulu	#40,d1
	add.w	d1,a1			; target y set

					; d0 already x/2
	add.w	d0,a1			; target x set
	;
	; clipping
	;
	; --> NO (for now)

	;
	; copy
	;
	move.w	(a0)+,d1		; w-1
	move.w	(a0),d3			; h-1
	move.w	#39,d4			; page stride: 40-w == 39-(w-1)
	sub.w	d1,d4

.rows	move.w	d1,d0
	moveq	#0,d6			; "had gfx character" flag
.cols	move.b	(a2)+,d5
	beq.b	.trans
	;
	; draw char -- need color setter before?
	;
	tst.w	d6
	bne.b	.color_check_done	; already had a char

;	cmp.w	d1,d0			; special case: first character in page row
;	bne.b	.not_x0
;	move.b	#'F',(a1)+
;	bra.b	.next

.not_x0	move.b	40*24-1(a1),d6
	cmp.b	d2,d6			; already the same color as desired?
	beq.b	.color_check_done
	move.b	d2,-1(a1)		; other x positions: set color setter in character before us

.color_check_done
	;
	; copy
	;
	move.b	d5,(a1)+
.next	st	d6 			; had gfx char = true
	dbf	d0,.cols
	add.w	d4,a1
	dbf	d3,.rows
	rts

.trans	;
	; no char (sprite is transparent here) -- need color reset?
	;
	tst.w	d6			; end of solid gfx char stride?
	beq.b	.no_color_reset
	move.b	40*24(a1),(a1)
.no_color_reset
	moveq	#0,d6			; had gfx char = false
	addq	#1,a1
	dbf	d0,.cols
	add.w	d4,a1
	dbf	d3,.rows
	rts

; ---------------------------------------------------------------------------
; draw_char_sprite(a0: sprite, d0: x, d1: y, d2: color setter command)
; ---------------------------------------------------------------------------
; needs current color buffer! --> update_renderbuffer_color_info
; color buffer is then kept current during draw

draw_char_sprite:
	SAVERESTORE d2-d7/a2-a4
	move.w	(a0)+,d3		; w
	move.w	(a0)+,d4		; h

	move.w	#40,d5			; dst stride
	sub.w	d3,d5
	moveq	#0,d6			; src stride

	; a0:src
	; a1:dst
	; d0:x
	; d1:y
	; d2:color
	; d3:w
	; d4:h
	; d5:dst stride
	; d6:src stride
	; d7:---

	;
	; clip top?
	;
	tst.w	d1
	bge.b	.nocliptop
	add.w	d1,d4			; new height
	bgt.b	.keeph
	rts
.keeph	move.w	d1,d7
	neg.w	d7
	mulu	d3,d7			; new src start (+w*clipped h)
	add.w	d7,a0
	moveq	#0,d1
.nocliptop
	;
	; clip bottom?
	;
	; example h=3
	;
	;  y	plot		y-24	y-24+h		new h
	; 19	19,20,21	-5	-2		3	--> no clip
	; 20	20,21,22	-4	-1		3	--> no clip
	; 21	21,22,23	-3	 0		3	--> no clip
	; 22	22,23		-2	 1		3-1
	; 23	23		-1	 2		3-2
	; 24			 0	 3		3-3	--> abort
	; 25			 1	 4		3-4	--> abort
	move.w	d1,d7
	sub.w	#24,d7		; y-24
	blt.b	.keepb
	rts
.keepb	add.w	d4,d7		; y-24+h
	ble.b	.noclipbottom
	sub.w	d7,d4		; new h
.noclipbottom
	;
	; clip left?
	;
	cmp.w	#1,d0
	bge.b	.noclipleft
	;
	; example w=5
	;
	;  x	tmpx	dst x	w	stride	dst stride
	;  3		3	5	0	40-5
	;  2		2	5	0	40-5
	;  1	 0	1	5	0	40-5
	;  0	-1	1	5-1=4	1	40-5+1
	; -1	-2	1	5-2=3	2	40-5+2
	; -2	-3	1	5-3=2	3	40-5+3
	; -3	-4	1	5-4=1	4	40-5+4
	; -4	-5	1	5-5=0	--> abort
	; -5	-6	1	5-6=-1	--> abort
	;
	subq	#2,d0			; tmpx
	add.w	d0,d3			; new w
	bgt.b	.keepw			; abort
	rts
.keepw	sub.w	d0,d5			; new dst stride
	sub.w	d0,d6			; new src stride
	sub.w	d0,a0			; new src start
	moveq	#1,d0			; new x
.noclipleft
	;
	; clip right?
	;
	cmp.w	#40,d0
	blt.b	.keepr
	rts
.keepr	move.w	d0,d7			; last
	add.w	d3,d7			; x+w
	sub.w	#40,d7
	tst.w	d7
	ble.b	.noclipright
	sub.w	d7,d3			; new w
	bne.b	.keepr2
	rts
.keepr2	add.w	d7,d5			; new dst stride
	add.w	d7,d6			; new src stride
.noclipright
	;
	; plot
	;
	subq	#1,d3			; plot w
	subq	#1,d4			; plot h
	lea	renderbuffer,a1
	mulu	#40,d1
	add.w	d1,a1
	add.w	d0,a1
.rows
	moveq	#0,d1			; "had gfx character" flag
	move.w	d3,d7
.cols	move.b	(a0)+,d0
	beq.b	.trans
	move.b	d2,40*24(a1)		; update color buffer
	move.b	d0,(a1)+		; set char
	tst.w	d1
	bne.b	.color_check_done	; already had a char
	move.b	40*24-2(a1),d0
	cmp.b	d2,d0			; already the same color as desired?
	beq.b	.color_check_done
	move.b	d2,-2(a1)		; other x positions: set color setter in character before us
	move.b	d2,-2+40*24(a1)		; update color buffer
.color_check_done
	st	d1
	dbf	d7,.cols
	add.w	d6,a0
	add.w	d5,a1
	dbf	d4,.rows
	rts

.trans
	tst.w	d1			; end of solid gfx char stride?
	beq.b	.no_color_reset
	move.b	40*24(a1),d0
	cmp.b	d2,d0			; already the same color as last used?
	beq.b	.no_color_reset
	move.b	d0,(a1)
.no_color_reset
	moveq	#0,d1			; had gfx char = false
	addq	#1,a1
	dbf	d7,.cols
	add.w	d6,a0
	add.w	d5,a1
	dbf	d4,.rows
	rts

; hack: if we draw onto land, invert the sixels
draw_char_sprite_with_bazooka_hack:
	SAVERESTORE d2-d7/a2-a4
	move.w	(a0)+,d3		; w
	move.w	(a0)+,d4		; h
	move.w	#40,d5			; dst stride
	sub.w	d3,d5
	moveq	#0,d6			; src stride
	tst.w	d1
	bge.b	.nocliptop
	add.w	d1,d4			; new height
	bgt.b	.keeph
	rts
.keeph	move.w	d1,d7
	neg.w	d7
	mulu	d3,d7			; new src start (+w*clipped h)
	add.w	d7,a0
	moveq	#0,d1
.nocliptop
	move.w	d1,d7
	sub.w	#24,d7		; y-24
	blt.b	.keepb
	rts
.keepb	add.w	d4,d7		; y-24+h
	ble.b	.noclipbottom
	sub.w	d7,d4		; new h
.noclipbottom
	cmp.w	#1,d0
	bge.b	.noclipleft
	subq	#2,d0			; tmpx
	add.w	d0,d3			; new w
	bgt.b	.keepw			; abort
	rts
.keepw	sub.w	d0,d5			; new dst stride
	sub.w	d0,d6			; new src stride
	sub.w	d0,a0			; new src start
	moveq	#1,d0			; new x
.noclipleft
	cmp.w	#40,d0
	blt.b	.keepr
	rts
.keepr	move.w	d0,d7			; last
	add.w	d3,d7			; x+w
	sub.w	#40,d7
	tst.w	d7
	ble.b	.noclipright
	sub.w	d7,d3			; new w
	bne.b	.keepr2
	rts
.keepr2	add.w	d7,d5			; new dst stride
	add.w	d7,d6			; new src stride
.noclipright
	subq	#1,d3			; plot w
	subq	#1,d4			; plot h
	lea	renderbuffer,a1
	mulu	#40,d1
	add.w	d1,a1
	add.w	d0,a1
.rows	moveq	#0,d1			; "had gfx character" flag
	move.w	d3,d7
.cols	move.b	(a0)+,d0
	beq.b	.trans
 cmp.b #SET_GFX_COL6_CYAN,40*24(a1)
 bne.b .notland1
 cmp.b #SET_GFX_COL6_CYAN,40*24-1(a1) ; also check character before to avoid color bleed (very crosshair-specific)
 bne.b .notland1
 btst #5,(a1) ; gfx char?
 beq.b .notland1
 move.w #SET_GFX_COL6_CYAN,d2
 ; sixel bits = 0100000 + F0EDCBA
 and.b #%1011111,d0
 eor.b d0,(a1)+
 bra.b .afterland1
.notland1
	move.b	d2,40*24(a1)		; update color buffer
	move.b	d0,(a1)+		; set char
.afterland1
	tst.w	d1
	bne.b	.color_check_done	; already had a char
	move.b	40*24-2(a1),d0
	cmp.b	d2,d0			; already the same color as desired?
	beq.b	.color_check_done
	move.b	d2,-2(a1)		; other x positions: set color setter in character before us
	move.b	d2,-2+40*24(a1)		; update color buffer
.color_check_done
	st	d1
	dbf	d7,.cols
	add.w	d6,a0
	add.w	d5,a1
	dbf	d4,.rows
	rts
.trans	tst.w	d1			; end of solid gfx char stride?
	beq.b	.no_color_reset
	move.b	40*24(a1),d0
	cmp.b	d2,d0			; already the same color as last used?
	beq.b	.no_color_reset
	move.b	d0,(a1)
.no_color_reset
	moveq	#0,d1			; had gfx char = false
	addq	#1,a1
	dbf	d7,.cols
	add.w	d6,a0
	add.w	d5,a1
	dbf	d4,.rows
	rts
